#include "generator.hpp"
#include "../include/hash.hpp"
#include <iostream>
#include <fstream>
#include <filesystem>
#include <sstream>
#include <vector>
#include "xxhash.hpp"
#include <sys/stat.h>  // For file permissions

namespace fs = std::filesystem;

static std::string sanitize(const std::string& name) {
    std::string out = name;
    for (char& c : out) if (!isalnum(c)) c = '_';
    return out;
}


static uint64_t fnv1a_64(const void* data, size_t len) {
    const uint8_t* p = static_cast<const uint8_t*>(data);
    uint64_t h = 0xcbf29ce484222325ULL;
    for (size_t i = 0; i < len; ++i)
        h = (h ^ p[i]) * 0x100000001b3ULL;
    return h;
}

// Helper function to output the _recreate_file_ utility function
static void output_recreate_file_utility(std::ofstream& cpp) {
    cpp << R"cpp(
// Utility function to recreate a file with proper permissions
static bool _recreate_file_(const std::filesystem::path& outpath, uint64_t file_hash, std::filesystem::perms file_perms, const unsigned char* filedata, size_t filedata_len) {
    namespace fs = std::filesystem;
    uint64_t existing_hash = 0;
    if (fs::exists(outpath)) {
        std::ifstream in(outpath, std::ios::binary);
        std::ostringstream oss;
        oss << in.rdbuf();
        std::string data = oss.str();
        existing_hash = fnv1a_64(data.data(), data.size());
    }
    bool needs_write = !fs::exists(outpath) || existing_hash != file_hash;
    if (needs_write) {
        fs::create_directories(outpath.parent_path());
        std::ofstream out(outpath, std::ios::binary);
        out.write(reinterpret_cast<const char*>(filedata), filedata_len);
        out.close();
        // Set the file permissions
        fs::permissions(outpath, file_perms);
    }
    
    if (!fs::exists(outpath)) {
        std::cout << "[dehydrate] " << outpath.filename() << ": created\n";
    } else if (needs_write) {
        std::cout << "[dehydrate] " << outpath.filename() << ": updated (hash changed)\n";
    }
    return needs_write;
}
)cpp";
}


void generate_file_code(const std::string& source, const std::string& destfolder, bool silent) {
    fs::path src(source);
    fs::path dest(destfolder);
    std::string ns = "recreate_" + sanitize(src.stem().string());
    std::string cppname = "_" + src.stem().string() + ".cpp";
    std::string hppname = "_" + src.stem().string() + ".hpp";
    fs::create_directories(dest);
    std::ifstream in(source, std::ios::binary);
    std::ostringstream oss;
    oss << in.rdbuf();
    std::string filedata = oss.str();
    uint64_t hash = fnv1a_64(filedata.data(), filedata.size());
    
    // Get source file permissions
    fs::perms src_perms = fs::status(src).permissions();
    
    // Write HPP
    std::ofstream hpp(dest / hppname);
    hpp << "#pragma once\n#include <string>\nnamespace " << ns << " {\nbool recreate_file(std::string destination_folder);\n}\n";
    // Write CPP
    std::ofstream cpp(dest / cppname);
    cpp << R"cpp(#include <fstream>
#include <filesystem>
#include <string>
#include <iostream>
#include <iomanip>

// Tiny dependency-free FNV-1a 64-bit hash
static uint64_t fnv1a_64(const void* data, size_t len) {
    const uint8_t* p = static_cast<const uint8_t*>(data);
    uint64_t h = 0xcbf29ce484222325ULL;
    for (size_t i = 0; i < len; ++i)
        h = (h ^ p[i]) * 0x100000001b3ULL;
    return h;
}
)cpp";
    cpp << "#include \"" << hppname << "\"\n";
    cpp << "namespace " << ns << " {\n";
    
    // Output the recreate_file utility function
    output_recreate_file_utility(cpp);
    
    // Write recreate_file function with embedded file data
    cpp << R"cpp(
bool recreate_file(std::string destination_folder) {
    namespace fs = std::filesystem;
    fs::path outpath = fs::path(destination_folder) / ")cpp" << src.filename().string() << R"cpp(";
    
    // File data embedded directly in the function
    const unsigned char filedata[] = {)cpp";
    
    // Embed file data directly
    for (size_t i = 0; i < filedata.size(); ++i) {
        if (i % 16 == 0) cpp << "\n        ";
        cpp << "0x" << std::hex << std::setw(2) << std::setfill('0') << (int)(unsigned char)filedata[i];
        if (i + 1 != filedata.size()) cpp << ", ";
    }
    
    cpp << std::dec << "\n    };\n";
    
    // Direct call to _recreate_file_ with hash and permissions defined inline
    cpp << "    return _recreate_file_(outpath, " << hash << "ULL, "
        << "std::filesystem::perms(" << static_cast<unsigned>(src_perms) << "), "
        << "filedata, " << filedata.size() << ");\n";
    
    cpp << "}\n";
    cpp << "}\n";
    
    if (!silent) {
        std::cout << "[dehydrate] Generated: " << (dest / cppname) << ", " << (dest / hppname) << std::endl;
    }
}

// Helper to recursively collect all files in a directory
template<typename F>
void walk_dir(const fs::path& dir, F&& f) {
    for (auto& p : fs::recursive_directory_iterator(dir)) {
        if (fs::is_regular_file(p)) f(p.path());
    }
}


void generate_folder_code(const std::string& source, const std::string& destfolder, bool silent) {
    fs::path src(source);
    fs::path dest(destfolder);
    std::string ns = "recreate_" + sanitize(src.stem().string());
    std::string cppname = "_" + src.stem().string() + ".cpp";
    std::string hppname = "_" + src.stem().string() + ".hpp";
    fs::create_directories(dest);
    // Collect all files
    std::vector<fs::path> files;
    walk_dir(src, [&](const fs::path& p) { files.push_back(p); });
    // Write HPP
    std::ofstream hpp(dest / hppname);

    // -------------------------------------------------------------------------
    // Generate HPP
    hpp << R"hpp(
#pragma once

/*

   THIS FILE IS AUTO-GENERATED BY DEHYDRATE.
   DO NOT EDIT THIS FILE.

*/


#include <string>
namespace )hpp" << ns << R"hpp( {
    bool recreate_tree(std::string destination_folder);
}
)hpp";


    // -------------------------------------------------------------------------
    // Write CPP
    std::ofstream cpp(dest / cppname);
    cpp << R"cpp(#include <fstream>
#include <filesystem>
#include <string>
#include <iostream>
#include <iomanip>



/*

   THIS FILE IS AUTO-GENERATED BY DEHYDRATE.
   DO NOT EDIT THIS FILE.

*/


)cpp";
    cpp << "#include \"" << hppname << "\"\n";
    cpp << "namespace " << ns << " {\n";

    cpp << R"cpp(

// Tiny dependency-free FNV-1a 64-bit hash
static uint64_t fnv1a_64(const void* data, size_t len) {
    const uint8_t* p = static_cast<const uint8_t*>(data);
    uint64_t h = 0xcbf29ce484222325ULL;
    for (size_t i = 0; i < len; ++i)
        h = (h ^ p[i]) * 0x100000001b3ULL;
    return h;
}

)cpp";

    // Output the recreate_file utility function
    output_recreate_file_utility(cpp);

    // Start writing recreate_tree - we'll embed file data directly in function
    cpp << R"cpp(
bool recreate_tree(std::string destination_folder) {
    namespace fs = std::filesystem;
    bool any_written = false;
)cpp";

    // Process each file
    for (const auto& file : files) {
        std::ifstream in(file, std::ios::binary);
        std::ostringstream oss;
        oss << in.rdbuf();
        std::string filedata = oss.str();
        uint64_t hash = fnv1a_64(filedata.data(), filedata.size());
        fs::perms file_perms = fs::status(file).permissions();
        std::string rel = fs::relative(file, src).string();
        std::string var = sanitize(rel);
        
        // Start a scope to limit data's lifetime
        cpp << "    {\n";
        cpp << "        // File: " << rel << "\n";
        cpp << "        fs::path outpath = fs::path(destination_folder) / \"" << rel << "\";\n";
        
        // Embed file data directly in the function
        cpp << "        const unsigned char filedata[] = {";
        for (size_t i = 0; i < filedata.size(); ++i) {
            if (i % 16 == 0) cpp << "\n            ";
            cpp << "0x" << std::hex << std::setw(2) << std::setfill('0') << (int)(unsigned char)filedata[i];
            if (i + 1 != filedata.size()) cpp << ", ";
        }
        cpp << "\n        };\n";
        // Reset to decimal format
        cpp << std::dec;
        
        // Direct call to _recreate_file_ with all parameters defined inline
        cpp << "        bool file_written = _recreate_file_(outpath, " 
             << hash << "ULL, std::filesystem::perms(" << static_cast<unsigned>(file_perms) << "), filedata, " << filedata.size() << ");\n";
        cpp << "        any_written = any_written || file_written;\n";
        cpp << "    }\n"; // Close scope to free memory
    }
    
    cpp << "    return any_written;\n";
    cpp << "}\n";
    cpp << "}\n";
    
    if (!silent) {
        std::cout << "[dehydrate] Generated: " << (dest / cppname) << ", " << (dest / hppname) << std::endl;
    }
} 